#include "client.h"

typedef struct {
	uint32_t id;
	char* ip;
	int port;
	auth_t* auth;
	pthread_t thrd;
	sock_t sock;
	SSL* ssl;
} serverconn_t;


bool isrunning;

sock_t clientsock;
serverconn_t* serverconns;

bool player;
uint32_t n;
uint32_t d;
uint32_t d1;

uint8_t* serverbuffer;
sem_t* serverinlocks;
sem_t* serveroutlocks;

uint8_t* phases;
uint8_t* recvbuffer1;
uint8_t* recvbuffer2;
uint8_t* recvbuffer3;
uint8_t* recvbuffer4;

uint32_t statSuccess;
uint32_t statFailed;

uint32_t hashlen = 32;

void runClient(bool isAlice) {
	printf("Running client as %s..\n", (isAlice ? "Alice" : "Bob"));
    mpoly_initialize(parms.numservers);

	isrunning = true;
	player = isAlice;
	n = parms.numservers;
	d = (n - 1) / 2;
	d1 = d + 1;

	// Connect clients and compute authentication
	clientsock = createClientConnection(isAlice);
	auth_t* auth = auth_vec_compute(clientsock, parms.numservers);
	printf("Connected to other client and computed authentication\n");


    // Spawn server threads and authenticate
	serverconns = malloc(parms.numservers * sizeof(serverconn_t));
	for(int i=0; i<parms.numservers; ++i) {
		serverconn_t* serverconn = &serverconns[i];
		serverconn->id = i;
		serverconn->ip = parms.ips[i];
		serverconn->port = parms.ports[i];
		serverconn->auth = auth;
		pthread_create(&(serverconn->thrd), NULL, createServerConnection, (void*)serverconn);
	}
	for(int i=0; i<parms.numservers; ++i)
		pthread_join(((serverconn_t)serverconns[i]).thrd, NULL);
	printf("Connected and authenticated to servers\n");


	// Spawn triple producer
	uint32_t batchlen = 2 * parms.numparallel * mf_p_bytes;
	uint32_t serverbufferlen = parms.numthread * parms.numservers * batchlen;
	serverbuffer = malloc(serverbufferlen);
	serverinlocks = malloc(parms.numthread * sizeof(sem_t));
	serveroutlocks = malloc(parms.numthread * sizeof(sem_t));
	for(int i=0; i<parms.numthread; ++i) {
		sem_init(&serverinlocks[i], 0, 1);
		sem_init(&serveroutlocks[i], 0, 0);
	}
	pthread_t produceThrd;
	pthread_create(&produceThrd, NULL, produceBatch, (void*)0);
	


	// Spawn processing threads and lock all
    int unique = (int)sfmt_genrand_uint32(&mf_sfmt_randstate);
    sync_init(unique, parms.numthread);	

    pthread_t* thrds = malloc(parms.numthread * sizeof(pthread_t));
	uint32_t* index = malloc(parms.numthread * sizeof(uint32_t));
	recvbuffer1 = malloc(parms.numthread * parms.numparallel * d * mf_p_bytes);
	recvbuffer2 = malloc(parms.numthread * mf_p_bytes);
	recvbuffer3 = malloc(parms.numthread * hashlen);
	recvbuffer4 = malloc(parms.numthread * parms.numparallel * 2 * mf_p_bytes);
	phases = malloc(parms.numthread);
	for(int i=0; i<parms.numthread; ++i) {
		index[i] = i;
		phases[i] = 0;
		sync_lock(i);
		pthread_create(&thrds[i], NULL, createOLEs, (void*)&index[i]);
	}

	pthread_t timerthrd;
	pthread_create(&timerthrd, NULL, timerThread, (void*)0);


    // Main thread receives data from the client
	int recvd;
    while(isrunning) {
        uint32_t index;
		recvd = mio_recv(clientsock, 4, (uint8_t*)&index);
		if(recvd == -1) {
			printf("Could not receive from other client\n");
			exit(1);
		}
        index = ntohl(index);
        if(index >= parms.numthread) printf("Invalid index!\n");
		if(phases[index] == 0) {
			recvd = mio_recv(clientsock, parms.numparallel * d * mf_p_bytes, &recvbuffer1[index * parms.numparallel * d * mf_p_bytes]);
			if(recvd == -1) {
				printf("Could not receive from other client 0\n");
				exit(1);
			}
			phases[index] = 1;
		}
		else if(phases[index] == 1) {
			recvd = mio_recv(clientsock, mf_p_bytes, &recvbuffer2[index * mf_p_bytes]);
			if(recvd == -1) {
				printf("Could not receive from other client 1\n");
				exit(1);
			}
			phases[index] = 2;
		}
		else if(phases[index] == 2) {
			recvd = mio_recv(clientsock, hashlen, &recvbuffer3[index * hashlen]);
			if(recvd == -1) {
				printf("Could not receive from other client 2\n");
				exit(1);
			}
			phases[index] = 0;
		}
		else if(phases[index] == 3) {
			recvd = mio_recv(clientsock, 2 * parms.numparallel * mf_p_bytes, &recvbuffer4[index * 2 * parms.numparallel * mf_p_bytes]);
			if(recvd == -1) {
				printf("Could not receive from other client 2\n");
				exit(1);
			}
			phases[index] = 0;
		}
        sync_unlock(index);
    }

	sync_destroy();
	
	mio_close(clientsock);
	mpoly_finalize();
}


void* timerThread(void* nada) {
	
	for(int i=3; i>=1; --i) {
		printf("Starting measurement in %d..\n", i);
		usleep(1000000);
	}

	watch_t startwatch;
	watch_start(&startwatch);

	int start = statSuccess;

	for(int i=0; i<28; ++i) {
		int before = statSuccess;
		usleep(1000000);
		int runs = statSuccess-before;
		int oles = runs * parms.numparallel;
		printf("After %d seconds: Generated %d OLEs over %d runs. Total status: %d good / %d bad\n", i+1, oles, runs, statSuccess, statFailed);
	}
	bool shouldstop = false;

	while(!shouldstop) {
		watch_t stopwatch;
		watch_start(&stopwatch);
		timespec_diff(&stopwatch, &startwatch, &stopwatch);
		double ms = watch_elapsedms(&stopwatch);
		if(ms >= 30000) {
			int num = statSuccess - start;

			printf("Results:\n  start = %d, num = %d\n  statSuccess = %d\n  statFailed = %d\n  totalOLEs = %d\n  averageOLE/s = %f\n  averageOLEms = %f\n  averagePROTms = %f\n",
			    start,
				num,
				statSuccess,
				statFailed,
				num*parms.numparallel,
				(num*parms.numparallel)/(ms/1000),
				ms / (num*parms.numparallel),
				ms / num);
			shouldstop = true;
			isrunning = false;
			exit(1);
		}
	}
     
	return 0;
}

void* produceBatch(void* nada) {
	uint32_t request = htonl(parms.numparallel);
	uint32_t batchlen = 2 * parms.numparallel * mf_p_bytes;

	int index = 0;
	while(index < parms.numthread) {
		// Wait for data receive available
		sem_wait(&serverinlocks[index]);

		// Receive data
		for(int i=0; i<parms.numservers; ++i) {
			serverconn_t* conn = (serverconn_t*)&serverconns[i];
			int sent = mio_ssend(conn->ssl, 4, (uint8_t*)&request);
			if(sent == -1) {
				printf("Could not send request to server!\n");
				exit(1);
			}
		}
		for(int i=0; i<parms.numservers; ++i) {
			serverconn_t* conn = (serverconn_t*)&serverconns[i];
			int recvd = mio_srecv(conn->ssl, batchlen, &serverbuffer[index * parms.numservers * batchlen + i*batchlen]);
			if(recvd == -1) {
				printf("Could not receive triples from server!\n");
				exit(1);
			}
		}

		// Unlock producer thread
		sem_post(&serveroutlocks[index]);

		// Proceed to serve next thread
		index++;
		if(index == parms.numthread) index = 0;
	}

	return 0;
}



void* createOLEs(void* index) {
	int id = *((int*)index);
	uint32_t nid = htonl(id);

	uint32_t batchlen = parms.numparallel * 2*mf_p_bytes;
	uint8_t* serverbuf = &serverbuffer[id * parms.numservers * batchlen];

	mf_t tmp; mf_init2(tmp);
	mf_vec_t aout;  mf_vec_init1(aout, parms.numparallel);
	mf_vec_t cout;  mf_vec_init1(cout, parms.numparallel);
	mf_vec_t a;		mf_vec_init1(a, parms.numparallel * n);
	mf_vec_t c;		mf_vec_init1(c, parms.numparallel * n);
	mf_vec_t aext;  mf_vec_init1(aext, parms.numparallel * d);
	mf_vec_t adiff; mf_vec_init1(adiff, parms.numparallel * d);
	mf_vec_t bdiff; mf_vec_init1(bdiff, parms.numparallel * d);
	mf_vec_t ares;mf_vec_init1(ares, parms.numparallel);
	mf_vec_t cres; mf_vec_init1(cres, parms.numparallel);
	mf_vec_t oares; mf_vec_init1(oares, parms.numparallel);
	mf_vec_t ocres; mf_vec_init1(ocres, parms.numparallel);

	uint32_t sendbufferlen = 4 + parms.numparallel * d * mf_p_bytes;
	uint32_t testbufferlen = 4 + mf_p_bytes;
	uint32_t commitbufferlen = 4 + SHA256_DIGEST_LENGTH;
	uint32_t checkbufferlen = 4 + parms.numparallel * 2 * mf_p_bytes;
	uint8_t* sendbuffer = malloc(sendbufferlen);
	uint8_t* testbuffer = malloc(testbufferlen);
	uint8_t* commitbuffer = malloc(commitbufferlen);
	uint8_t* checkbuffer = malloc(checkbufferlen);
	memcpy(sendbuffer, &nid, 4);
	memcpy(testbuffer, &nid, 4);
	memcpy(commitbuffer, &nid, 4);
	memcpy(checkbuffer, &nid, 4);

	mprg_t* prg = malloc(sizeof(mprg_t));
	mprg_init(prg);
	
	int sent;

	mf_t minpoint;
	mf_t testpoint;
	mf_t otestpoint;
	mf_init1(minpoint);
	mf_init1(testpoint);
	mf_init1(otestpoint);
	mf_set_ui(minpoint, parms.numservers+1);
	
	uint8_t* recvbuf1 = &recvbuffer1[id * parms.numparallel * d * mf_p_bytes];
	uint8_t* recvbuf2 = &recvbuffer2[id * mf_p_bytes];
	uint8_t* recvbuf3 = &recvbuffer3[id * hashlen];
	uint8_t* recvbuf4 = &recvbuffer3[id * 2 * parms.numparallel * mf_p_bytes];

	while(true) {
		// Get current data and unlock producer
		sem_wait(&serveroutlocks[id]);

		//watch_t startwatch;
		//watch_start(&startwatch);

		for(int i=0; i<parms.numparallel; ++i) {
			for(int j=0; j<n; ++j) {
				mf_import(a[i*n+j], &serverbuf[j*batchlen+i*mf_p_bytes]);
				mf_import(c[i*n+j], &serverbuf[j*batchlen+i*mf_p_bytes+parms.numparallel*mf_p_bytes]);
				if(mf_iszero(a[i*n+j]) || mf_iszero(c[i*n+j])) {
					printf("Received zero? wtf!\n");
					printHex(&serverbuffer[j*batchlen+i*mf_p_bytes], 2*parms.numparallel*mf_p_bytes);
					exit(1);
				}
			}
			//gmp_printf("Polynomial A: %Zd %Zd %Zd %Zd %Zd\n", a[i*n+0], a[i*n+1], a[i*n+2], a[i*n+3], a[i*n+4]);
			//gmp_printf("Polynomial C: %Zd %Zd %Zd %Zd %Zd\n", c[i*n+0], c[i*n+1], c[i*n+2], c[i*n+3], c[i*n+4]);

		}
		sem_post(&serverinlocks[id]);


		// Compute A(x) and send first message
		for(int i=0; i<parms.numparallel; ++i) {
			mpoly_computePoints(&aext[i*d], aout[i], &a[i*n]);
			mf_vec_sub(&adiff[i*d], &aext[i*d], &a[i*n+d1], d);
			//gmp_printf("Extended A (%d) %Zd: %Zd %Zd\n", i*d, aout[i], aext[i*d], aext[i*d+1]);
			//gmp_printf("Diffed   A: %Zd %Zd\n", adiff[i*d], adiff[i*d+1]);

			// Append to message
			for(int j=0; j<d; ++j) {
				mf_export(&sendbuffer[4 + i*d*mf_p_bytes + j*mf_p_bytes], adiff[i*d+j]);
			}
		}
		sync_slock();
		sent = mio_send(clientsock, sendbufferlen, sendbuffer);
		if(sent == -1) {
			printf("Could not send!\n");
			exit(1);
		}
		sync_sunlock();


		// Receive other partys masked input
		sync_lock(id);
		mf_vec_import(bdiff, recvbuf1, parms.numparallel * d);
		// for(int i=0; i<parms.numparallel; ++i) {
		// 	gmp_printf("Diffed B: %Zd %Zd\n", bdiff[i*d], bdiff[i*d+1]);
		// }

		// Recover my share of C(x)
		for(int i=0; i<parms.numparallel; ++i) {
			if(player == 1) { // Alice
				for(int j=0; j<d; ++j) {
					mf_mul(tmp, a[i*n+d1+j], bdiff[i*d + j]);
					mf_add(c[i*n+d1+j], c[i*n+d1+j], tmp);
					//gmp_printf("Alice c%d += a*e = %Zd * %Zd = %Zd\n", d1+j, a[i*n+d1+j], bdiff[i*d+j], c[i*n+d1+j]);
				}
			}
			else { // Bob
				for(int j=0; j<d; ++j) {
					mf_mul(tmp, aext[i*d+j], bdiff[i*d + j]);
					mf_add(c[i*n+d1+j], c[i*n+d1+j], tmp);
					//gmp_printf("Bob c%d += a*e = %Zd * %Zd = %Zd\n", d1+j, aext[i*d+j], bdiff[i*d+j], c[i*n+d1+j]);
				}
			}
			//gmp_printf("Polynomial C: %Zd %Zd %Zd %Zd %Zd\n", c[i*n+0], c[i*n+1], c[i*n+2], c[i*n+3], c[i*n+4]);
			mpoly_computeAtZero(cout[i], &c[i*n]);
			// if(player == 1) {
			// 	gmp_printf("Corrected%d %d*%d: a=%Zd; c1=%Zd\n", id, i, d, aout[i], cout[i]);
			// }
			// else {
			// 	gmp_printf("Corrected%d %d*%d: b=%Zd; c2=%Zd\n", id, i, d, aout[i], cout[i]);
			// }
		}




		// Sample polynomial test point
		do { mf_srand(testpoint, prg); } while(mpz_cmp(testpoint, minpoint) < 0);
		mf_export(&testbuffer[4], testpoint);


		// Send test point
		sync_slock();
		sent = mio_send(clientsock, testbufferlen, testbuffer);
		if(sent == -1) {
			printf("Could not send!\n");
			exit(1);
		}
		sync_sunlock();

		// Receive others test point
		sync_lock(id);
		mf_import(otestpoint, recvbuf2);



		// Compute and send committed response
		for(int i=0; i<parms.numparallel; ++i) {
			mpoly_computePointA(ares[i], otestpoint, &a[i*n]);
			mf_export(&checkbuffer[4 + i*mf_p_bytes], ares[i]);

			mpoly_computePointC(cres[i], otestpoint, &c[i*n]);
			mf_export(&checkbuffer[4 + parms.numparallel * mf_p_bytes + i*mf_p_bytes], cres[i]);
		}

		SHA256_CTX ctx;
		SHA256_Init(&ctx);
		SHA256_Update(&ctx, &checkbuffer[4], parms.numparallel * mf_p_bytes * 2);
		SHA256_Final(&commitbuffer[4], &ctx);

		sync_slock();
		sent = mio_send(clientsock, commitbufferlen, commitbuffer);
		if(sent == -1) {
			printf("Could not send!\n");
			exit(1);
		}
		sync_sunlock();

		// Receive other party's commited response to recvbuf3
		sync_lock(id);

		// Send own response to challenge
		sync_slock();
		sent = mio_send(clientsock, checkbufferlen, checkbuffer);
		if(sent == -1) {
			printf("Could not send!\n");
			exit(1);
		}
		sync_sunlock();

		// Receive other party's commitment response
		sync_lock(id);


		// Check his response was correctly committed
		uint8_t realhash[hashlen];
		SHA256_Init(&ctx);
		SHA256_Update(&ctx, &recvbuf4[4], parms.numparallel * mf_p_bytes * 2);
		SHA256_Final(realhash, &ctx);

		if(memcmp(recvbuf3, realhash, hashlen) != 0) {
			printf("Commitment failed\n");
			exit(1);
		}

		// Compute check value
		mf_vec_import(oares, &recvbuf4[0], parms.numparallel);
		mf_vec_import(ocres, &recvbuf4[parms.numparallel * mf_p_bytes], parms.numparallel);

		mf_vec_zero(ares, parms.numparallel);
		mf_vec_zero(cres, parms.numparallel);
		for(int i=0; i<parms.numparallel; ++i) {
			mpoly_computePointA(ares[i], testpoint, &a[i*n]);
			mpoly_computePointC(cres[i], testpoint, &c[i*n]);
		}

		bool valid = true;	
		for(int i=0; i<parms.numparallel; ++i) {
			mf_mul(tmp, oares[i], ares[i]);
			mf_sub(tmp, tmp, ocres[i]);
			if(mpz_cmp(tmp, cres[i]) != 0) {
				valid = false;
				gmp_printf("Invalid %d :( %Zd * %Zd != %Zd + %Zd\n", id, ares[i], oares[i], cres[i], ocres[i]);
				break;
				//exit(1); // arh måske lige i overkanten
			}
		}
		
		if(valid) {
			sync_slock();
			++statSuccess;
			sync_sunlock();
		}
		else {
			sync_slock();
			++statFailed;
			sync_sunlock();
		}
/*


		//gmp_printf("My testpoint: %Zd\n", testpoint); printf("\n");
		//gmp_printf("Others testpoint: %Zd ", otestpoint); printf("\n");




		watch_t stopwatch;
		watch_start(&stopwatch);
		timespec_diff(&stopwatch, &startwatch, &stopwatch);
		double ms = watch_elapsedms(&stopwatch);
		printf("Generated %d OLEs in %f ms\n", statSuccess * parms.numparallel, ms);
*/

		// Reset
		mf_vec_zero(aout, parms.numparallel);
		mf_vec_zero(cout, parms.numparallel);
		mf_vec_zero(aext, parms.numparallel * d);
		mf_vec_zero(ares, parms.numparallel);
		mf_vec_zero(cres, parms.numparallel);

		//printf("Success: %d, failed: %d\n", statSuccess, statFailed);
	}


	return 0;
}




void* createServerConnection(void* c) {
	serverconn_t* conn = (serverconn_t*)c;
	size_t datalen = 4 + mf_p_bytes + auth_export_bytes;
	uint8_t* data = malloc(datalen);

	// Set requested field	
	uint32_t l = htonl((uint32_t)mf_p_bytes);
	memcpy(&data[0], &l, 4);
	mf_export(&data[4], mf_p);

	// Compute authentication
	auth_t auth = *(conn->auth);
	auth_export(&data[4 + mf_p_bytes], auth);
	
	// Connect to server
	printf("Connecting to %s : %d\n", conn->ip, conn->port);
	sockaddrp_t addr = malloc(sizeof(sockaddr_t));
	mio_parseaddr(addr, conn->ip, conn->port);	
	conn->sock = mio_sconnect(&(conn->ssl), addr);
	mio_ssend(conn->ssl, datalen, data);
	free(data);

	return 0;
}


sock_t createClientConnection(bool isAlice) {
	sockaddrp_t addr = malloc(sizeof(sockaddr_t));
	sock_t sock;
	if(isAlice) {		
		// Wait for connection
		printf("Waiting for connections...\n");
		sock_t server = mio_listen(parms.port);
		memset(addr, 0, sizeof(sockaddr_t));
		sock = mio_accept(server, addr);
		char *clientip = inet_ntoa(addr->sin_addr);
		printf("New connection from Bob at %s:%d\n", clientip, addr->sin_port);
		mio_close(server);
		mio_check(sock);
		printf("Connection established.\n");
	}
	else {
		// Create connections
		printf("Connecting to Alice..\n");
		mio_parseaddr(addr, parms.ip, parms.port);	
		sock = mio_connect(addr);
		mio_check(sock);		
		printf("Connection established\n");
	}
	free(addr);
	return sock;
}