#include "field.h"


///////////////////////////////////////////
// State
///////////////////////////////////////////

mf_t mf_p;
size_t mf_p_bits;
size_t mf_p_bytes;

mf_t mf_zero;
mf_t mf_one;
mf_t mf_randmax;

size_t randbufferlen;
size_t randbufferbytes;
uint64_t randbuffermask;




///////////////////////////////////////
// Functions
///////////////////////////////////////

void mf_initialize(mf_t p) {
#if (FINITEFIELD == FF31) || (FINITEFIELD == FF63)
	mf_setfield(p);
	mf_zero = 0;
	mf_one = 1;
#elif FINITEFIELD == FFGMP
	mpz_init(mf_p);
	mf_setfield(p);
	mpz_init(mf_zero);
	mpz_init(mf_one);
	mpz_set_ui(mf_one, 1);
#endif
}

void mf_setfield(mf_t p) {
#if (FINITEFIELD == FF31) || (FINITEFIELD == FF63)

	mf_p = p;
	mf_p_bits = 0; mf_t tmp = mf_p; while(tmp >>= 1) ++mf_p_bits;	
	mf_p_bytes = (mf_p_bits + 7) / 8;
	uint64_t bitsToClear = 64 - mf_p_bits;
	randbuffermask = ((uint64_t)1 << mf_p_bits) - (uint64_t)1;

#elif FINITEFIELD == FFGMP

	mpz_set(mf_p, p);
  	mf_p_bits = mpz_sizeinbase(mf_p, 2);
	mf_p_bytes = (mf_p_bits + 7) / 8;

	randbufferbytes = mf_p_bytes;
	while((randbufferbytes % 8) != 0) ++randbufferbytes;
	randbufferlen = randbufferbytes / 8;

	if((mf_p_bits % 64) == 0) randbuffermask = ~0;
	else randbuffermask = ((uint64_t)1 << (uint64_t)(mf_p_bits%64)) - (uint64_t)1;

#endif
}


void mf_finalize() {
	mf_p_bits = 0;
	mf_p_bytes = 0;
#if FINITEFIELD == FFGMP
	mpz_clear(mf_p);
	mpz_clear(mf_zero);
	mpz_clear(mf_one);
#endif
}



void mf_print(mf_t x) {
#if (FINITEFIELD == FF31)
	printf("%ld", x);
#elif (FINITEFIELD == FF63) 
	printf("%lld", x);
#else
	gmp_printf("%Zd", x);
#endif
}


// Note: not safe across different endianess for optimized fields
void mf_import(mf_t r, uint8_t* a) {
#if (FINITEFIELD == FF31)
	r = *((mf_t*)(a));
#elif (FINITEFIELD == FF63) 
	r = *((mf_t*)(a));
#else
	mpz_import(r, 1, 1, mf_p_bytes, 0, 0, a);
#endif
}

// Note: not safe across different endianess for optimized fields
void mf_export(uint8_t* r, mf_t a) {
#if (FINITEFIELD == FF31)
	memcpy(r, &a, mf_p_bytes);
#elif (FINITEFIELD == FF63) 
	memcpy(r, &a, mf_p_bytes);
#else
	size_t written = 0;
	mpz_export(r, &written, 1, mf_p_bytes, 0, 0, a);
	if(written == 0) {
		memset(r, 0, mf_p_bytes);
	}
#endif
}


void mf_vec_import(mf_vec_t r, uint8_t* a, size_t l) {
	int p = 0;
	for(int i=0; i<l; ++i) {
		mf_import(r[i], &a[p]);
		p += mf_p_bytes;
	}
}

void mf_vec_export(uint8_t* r, mf_vec_t a, size_t l) {
	int p = 0;
	for(int i=0; i<l; ++i) {
		mf_export(&r[p], a[i]);
		p += mf_p_bytes;
	}
}













// Shamelessly borrowed from di-mgt.com, extended euclidean
#if (FINITEFIELD != FFGMP)
void mf_inv(mf_t* r, mf_t u) {
  unsigned long long inv, u1, u3, v1, v3, t1, t3, q;
  int iter;
  u1 = 1;
  u3 = (unsigned long long)u;
  v1 = 0;
  v3 = (unsigned long long)mf_p;
  iter = 1;
  while (v3 != 0)
  {
      q = u3 / v3;
      t3 = u3 % v3;
      t1 = u1 + q * v1;
      u1 = v1; v1 = t1; u3 = v3; v3 = t3;
      iter = -iter;1
  }
  if (u3 != 1) *r = 0;   /* Error: No inverse exists */
  if (iter < 0) inv = mf_p - (mf_t)u1;
  else inv = (mf_t)u1;
  *r = inv;
}
#endif




#if (FINITEFIELD == FFGMP)
void mf_srand(mf_t r, mprg_t* prg) {
	uint64_t* randbuffer = malloc(randbufferbytes);
	do {
		for(size_t i=0; i<randbufferlen; ++i) {
			randbuffer[i] = mprg_rand64(prg);
		}
		randbuffer[randbufferlen-1] &= randbuffermask;
		mpz_import(r, randbufferlen, 1, 8, -1, 0, randbuffer);
	}
	while(mpz_cmp(r, mf_p) >= 0);
	free(randbuffer);
}
#else
void mf_srand(mf_t* r, mprg_t* prg) {
	do {
#if FINITEFIELD == FF31
		f_t r = mprg_rand32(prg) & (uint32_t)randbuffermask;
#else
		f_t r = mprg_rand64(prg) & randbuffermask;
#endif
	}
	while(r >= mf_p);
}
#endif


#if (FINITEFIELD != FFGMP)
void mf_pow(mf_t* r, mf_t a, mf_t b) {
    r = 1; 
    while (y > 0) {
        if (y & 1) r = (r*a) % p;
        b = b>>1;
        a = (a*a) % p;  
    }
}
#endif
