#include "polynomials.h"

int n;
int d;
int d1;
mf_vec_t basis;
mf_vec_t delta;
mf_vec_t lambda;
mf_vec_t delta2;
mf_vec_t lambda2;
mf_vec_t theta;
mf_vec_t theta2;

void mpoly_initialize(int num) {
    n = num;
    d = (n - 1) / 2;
    d1 = d + 1;

    mf_t tmp, tmp2, xi, xj;
    mf_init2(tmp);
    mf_init2(tmp2);
    mf_init2(xi);
    mf_init2(xj);

    // Define basis[i]=i for i=0,n
    mf_vec_init1(basis, n);
    mf_set(basis[0], mf_one);
    for(int i=1; i<n; ++i) {
        mf_add(basis[i], basis[i-1], mf_one);
    }

    // Compute delta and lambda, d=(xi-xj)^-1, l=xj*d for deg. d
    mf_vec_init1(delta, d1 * d1);
    mf_vec_init1(lambda, d1 * d1);
    mf_vec_init(theta, d1);
    for(int i=0; i<d1; ++i) {
        mf_set(tmp, mf_one);
        for(int j=0; j<d1; ++j) {
            if(j == i) continue;
            mf_set(xi, basis[i]);
            mf_set(xj, basis[j]);
            
            mf_sub(delta[i * d1 + j], xi, xj);
            mf_inv(delta[i * d1 + j], delta[i * d1 + j]);

            mf_mul(lambda[i * d1 + j], delta[i * d1 + j], xj);
            mf_sub(lambda[i * d1 + j], mf_zero, lambda[i * d1 + j]);
	        mf_mul(tmp, tmp, lambda[i * d1 + j]);
        }
	    mf_set(theta[i], tmp);
    }

    // Compute delta, lambda, theta for degre n-1
    mf_vec_init1(delta2, n * n);
    mf_vec_init1(lambda2, n * n);
    mf_vec_init1(theta2, n);
    for(int i=0; i<n; ++i) {
        mf_set(tmp, mf_one);
        for(int j=0; j<n; ++j) {
            if(j == i) continue;
            mf_set(xi, basis[i]);
            mf_set(xj, basis[j]);
            
            mf_sub(delta2[i * n + j], xi, xj);
            mf_inv(delta2[i * n + j], delta2[i * n + j]);

            mf_mul(lambda2[i * n + j], delta2[i * n + j], xj);
            mf_sub(lambda2[i * n + j], mf_zero, lambda2[i * n + j]);

	        mf_mul(tmp, tmp, lambda2[i * n + j]);
        }
	    mf_set(theta2[i], tmp);
    }

    mf_clear(tmp);
    mf_clear(tmp2);
    mf_clear(xi);
    mf_clear(xj);
}


void mpoly_finalize() {
    mf_vec_clear(basis, n);

    mf_vec_clear(delta, d1 * d1);
    mf_vec_clear(lambda, d1 * d1);
    mf_vec_clear(theta, d1);

    mf_vec_clear(delta2, n * n);
    mf_vec_clear(lambda2, n * n);
    mf_vec_clear(theta2, n);
}


void mpoly_computePoints(mf_t* r, mf_t z, mf_t* a) {
    // From d+1 points at 0,..,d on input a
    // Output to r with points d+1,..,n-1=2d
    // Output to z with f(0)
    mf_t tmp;
    mf_t tmp2;
    mf_init2(tmp);
    mf_init2(tmp2);
    for(int p=0; p<d; ++p) {
        for(int i=0; i<=d; ++i) {
            mf_set(tmp2, mf_one);
            for(int j=0; j<=d; ++j) {
                if(j == i) continue;
                mf_mul(tmp, basis[d1 + p], delta[i * d1 + j]);
                mf_add(tmp, tmp, lambda[i * d1 + j]);
                mf_mul(tmp2, tmp2, tmp);
            }
            mf_mul(tmp, tmp2, a[i]);
            mf_add(r[p], r[p], tmp);
        }
    }

    for(int i=0; i<=d; ++i) {
        mf_mul(tmp, theta[i], a[i]);
        mf_add(z, z, tmp);
    }

    mf_clear(tmp);
    mf_clear(tmp2);
}

void mpoly_computePointA(mf_t r, mf_t x, mf_t* a) {
    mf_t tmp;
    mf_t tmp2;
    mf_init2(tmp);
    mf_init2(tmp2);
    for(int i=0; i<d1; ++i) {
        mf_set(tmp2, mf_one);
        for(int j=0; j<d1; ++j) {
            if(j == i) continue;
            mf_mul(tmp, x, delta[i * d1 + j]);
            mf_add(tmp, tmp, lambda[i * d1 + j]);
            mf_mul(tmp2, tmp2, tmp);
        }
        mf_mul(tmp, tmp2, a[i]);
        mf_add(r, r, tmp);
    }
    mf_clear(tmp);
    mf_clear(tmp2);
}

void mpoly_computePointC(mf_t r, mf_t x, mf_t* c) {
    mf_t tmp;
    mf_t tmp2;
    mf_init2(tmp);
    mf_init2(tmp2);
    for(int i=0; i<n; ++i) {
        mf_set(tmp2, mf_one);
        for(int j=0; j<n; ++j) {
            if(j == i) continue;
            mf_mul(tmp, x, delta2[i * n + j]);
            mf_add(tmp, tmp, lambda2[i * n + j]);
            mf_mul(tmp2, tmp2, tmp);
        }
        mf_mul(tmp, tmp2, c[i]);
        mf_add(r, r, tmp);
    }
    mf_clear(tmp);
    mf_clear(tmp2);
}

void mpoly_computeAtZero(mf_t r, mf_t* c) {
    // From n=2d+1 points
    mf_t tmp;
    mf_init2(tmp);
    for(int i=0; i<n; ++i) {
        mf_mul(tmp, theta2[i], c[i]);
        mf_add(r, r, tmp);
    }
    mf_clear(tmp);
}
