#include "prg.h"


void mprg_init(mprg_t* prg) {
	char s[128];
	int rfp = open("/dev/urandom", O_RDONLY); // This should be random, but is blocking :-(
	if (rfp < 0) {
		printf("Could not open /dev/random for seeding PRG\n");
		exit(1);
	}
	else {
		ssize_t res = read(rfp, s, sizeof(s));
		if (res < 0) {
			printf("Could not read /dev/random for seeding PRG\n");
			exit(1);
		}
    	close(rfp);
	}
    prg->seed[0] = _mm_loadu_si128((__m128i const*)s);
    aes_expandkey(prg->seed);
    prg->state = 0;
	prg->goodBytes = 0;
}

void mprg_seed(mprg_t* prg, uint8_t* s) {
    prg->seed[0] = _mm_loadu_si128((__m128i const*)s);
    aes_expandkey(prg->seed);
    prg->state = 0;
	prg->goodBytes = 0;
}

__uint128_t mprg_rand128(mprg_t* prg) {
    prg->randomness = _mm_cvtsi64_si128(prg->state);
    prg->randomness = aes_encrypt(prg->randomness, prg->seed);
    prg->state += 1;
    prg->goodBytes = 0;
    return (__uint128_t)prg->randomness;
}

uint64_t mprg_rand64(mprg_t* prg) {
    uint64_t res;
    if(prg->goodBytes == 8) {
        prg->randomness = _mm_bsrli_si128(prg->randomness, 8);
        prg->goodBytes = 0;
        res = (uint64_t)_mm_cvtsi128_si64(prg->randomness);
    } else {
        prg->randomness = _mm_cvtsi64_si128(prg->state);
        prg->randomness = aes_encrypt(prg->randomness, prg->seed);
    	prg->state += 1;
        prg->goodBytes = 8;
        res = (uint64_t)_mm_cvtsi128_si64(prg->randomness);
    }
    return res;
}

uint32_t mprg_rand32(mprg_t* prg) {
    uint32_t res;
    if(prg->goodBytes >= 4) {
        prg->randomness = _mm_bsrli_si128(prg->randomness, 4);
        prg->goodBytes -= 4;
        res = (uint32_t)_mm_cvtsi128_si32(prg->randomness);
    } else {
        prg->randomness = _mm_cvtsi64_si128(prg->state);
        prg->randomness = aes_encrypt(prg->randomness, prg->seed);
    		prg->state += 1;
        prg->goodBytes = 12;
        res = (uint64_t)_mm_cvtsi128_si64(prg->randomness);
    }
    return res;
}

void mprg_test(mprg_t* prg) {
	/*
	uint8_t* myseed = malloc(16);
	memset(myseed, 0, 16);
	mprg_seed(myseed);

	printf("Correct AES eval of key zero:\n");
	printf("  For message 0: 66E94BD4EF8A2C3B884CFA59CA342B2E\n");
	printf("  For message 1: 47711816E91D6FF059BBBF2BF58E0FD3\n");

	__uint128_t v3 = 0;
	uint8_t* r = (uint8_t*)&v3;
	v3 = mprg_rand128();
	printf("Random128:"); printHex(r, 16); printf("\n");
	v3 = mprg_rand128();
	printf("Random128:"); printHex(r, 16); printf("\n");

	mprg_seed(myseed);
	uint64_t v = mprg_rand64();
	r = (uint8_t*)&v;
	printf("Random64:"); printHex(r, 8); printf("\n");
	v = mprg_rand64();
	printf("Random64:"); printHex(r, 8); printf("\n");
	v = mprg_rand64();
	printf("Random64:"); printHex(r, 8); printf("\n");
	v = mprg_rand64();
	printf("Random64:"); printHex(r, 8); printf("\n");

	mprg_seed(myseed);
	uint32_t v2 = 0;
	r = (uint8_t*)&v2;
	v2 = mprg_rand32();
	printf("Random32:"); printHex(r, 4); printf("\n");
	v2 = mprg_rand32();
	printf("Random32:"); printHex(r, 4); printf("\n");
	v2 = mprg_rand32();
	printf("Random32:"); printHex(r, 4); printf("\n");
	v2 = mprg_rand32();
	printf("Random32:"); printHex(r, 4); printf("\n");
	v2 = mprg_rand32();
	printf("Random32:"); printHex(r, 4); printf("\n");
	v2 = mprg_rand32();
	printf("Random32:"); printHex(r, 4); printf("\n");
	v2 = mprg_rand32();
	printf("Random32:"); printHex(r, 4); printf("\n");
	v2 = mprg_rand32();
	printf("Random32:"); printHex(r, 4); printf("\n");
	*/
}
