#include "sync.h"

static int UNIQUE;
static int NUMLOCKS;
#ifdef WIN32
static HANDLE* locks;
static HANDLE lock;
#else
static sem_t* locks;
static pthread_mutex_t lock;
#endif



void sync_init(int unique, int num) {
	UNIQUE = unique;
	NUMLOCKS = num;
#ifdef WIN32
	lock = CreateMutex(NULL, FALSE, NULL);
	locks = malloc(NUMLOCKS * sizeof(HANDLE));
	for (int i = 0; i < NUMLOCKS; i++) {
		locks[i] = CreateSemaphore(NULL, 1, 1, NULL);
	}
#else
	pthread_mutex_init(&lock, NULL);
	locks = malloc(NUMLOCKS * sizeof(sem_t));
	for (int i = 0; i < NUMLOCKS; i++) {
		sem_init(&locks[i], 0, 1);
	}
#endif
}


void sync_slock() {
#ifdef WIN32
	WaitForSingleObject(lock, INFINITE);
#else
	pthread_mutex_lock(&lock);
#endif
}

void sync_sunlock() {
#ifdef WIN32
	ReleaseMutex(lock);
#else
	pthread_mutex_unlock(&lock);
#endif
}

void sync_lock(int index) {
#ifdef WIN32
	DWORD lol = WaitForSingleObject(locks[index], INFINITE);
#else
	sem_wait(&locks[index]);
#endif
}

void sync_unlock(int index) {
#ifdef WIN32
	//ReleaseMutex(locks[index]);
	ReleaseSemaphore(locks[index], 1, NULL);
#else
	sem_post(&locks[index]);
#endif
}

void sync_destroy() {
#ifdef WIN32
	for (int i = 0; i < NUMLOCKS; i++) {
		CloseHandle(locks[i]);
	}
	CloseHandle(lock);
#else
	for (int i = 0; i < NUMLOCKS; i++) {
		sem_destroy(&locks[i]);
	}
	pthread_mutex_destroy(&lock);
#endif
	free(locks);
}
